@@ -121,18 +121,39 @@ def _wrap_in_dbus_variant(value):
     raise dbus.exceptions.DBusException(f'could not wrap type {type(value)}')
 
 
+def _convert_args(signature: str, args: Tuple[Any, ...]) -> List[Any]:
+    """
+    Convert types of arguments according to signature, using
+    MethodCallMessage.append(); this will also provide type/length
+    checks, except for the case of an empty signature
+    """
+    try:
+        if signature == '' and len(args) > 0:
+            raise TypeError('Fewer items found in D-Bus signature than in Python arguments')
+        m = dbus.connection.MethodCallMessage('a.b', '/a', 'a.b', 'a')
+        m.append(signature=signature, *args)
+        return m.get_args_list()
+    except Exception as e:
+        raise dbus.exceptions.DBusException(f'Invalid arguments: {str(e)}',
+                                            name='org.freedesktop.DBus.Error.InvalidArgs')
+
+
 def loggedmethod(self, func):
     """Decorator for a method to end in the call log"""
 
     @functools.wraps(func)
     def wrapper(*args, **kwargs):
         fname = func.__name__
-        log_args = args[1:]  # args[0] is self
-        self.log(fname + _format_args(log_args))
-        self.call_log.append((int(time.time()), fname, log_args))
-        self.MethodCalled(fname, log_args)
+        self_arg, args = args[0], args[1:]
 
-        return func(*args, **kwargs)
+        in_signature = getattr(func, '_dbus_in_signature', '')
+        args = _convert_args(in_signature, args)
+
+        self.log(fname + _format_args(args))
+        self.call_log.append((int(time.time()), fname, args))
+        self.MethodCalled(fname, args)
+
+        return func(*[self_arg, *args], **kwargs)
 
     return wrapper
 
@@ -191,8 +212,11 @@ class DBusMockObject(dbus.service.Object):  # pylint: disable=too-many-instance-
         self._reset(props)
 
     def __del__(self) -> None:
-        if self.logfile and self.is_logfile_owner:
-            self.logfile.close()
+        try:
+            if self.logfile and self.is_logfile_owner:
+                self.logfile.close()
+        except AttributeError:
+            pass
 
     def _set_up_object_manager(self) -> None:
         '''Set up this mock object as a D-Bus ObjectManager.'''
@@ -583,10 +607,28 @@ class DBusMockObject(dbus.service.Object):  # pylint: disable=too-many-instance-
         self._template = template
         self._template_parameters = parameters
 
+    def _emit_signal(self, interface: str, name: str, signature: str, sigargs: Tuple[Any, ...], details: PropsType) -> None:
+        # pylint: disable=protected-access
+        if not interface:
+            interface = self.interface
+
+        args = _convert_args(signature, sigargs)
+
+        sig = dbus.lowlevel.SignalMessage(self.path, interface, name)
+        sig.append(*args, signature=signature)
+        dest = details.get("destination", None)
+        if dest is not None:
+            sig.set_destination(dest)
+
+        for location in self.locations:
+            conn = location[0]
+            conn.send_message(sig)
+        self.log(f'emit {self.path} {interface}.{name}{_format_args(args)}')
+
     @dbus.service.method(MOCK_IFACE,
                          in_signature='sssav',
                          out_signature='')
-    def EmitSignal(self, interface: str, name: str, signature: str, args: List[Any]) -> None:
+    def EmitSignal(self, interface: str, name: str, signature: str, sigargs: Tuple[Any, ...]) -> None:
         '''Emit a signal from the object.
 
         interface: D-Bus interface to send the signal from. For convenience you
@@ -599,26 +641,27 @@ class DBusMockObject(dbus.service.Object):  # pylint: disable=too-many-instance-
         args: variant array with signal arguments; must match order and type in
               "signature"
         '''
-        # pylint: disable=protected-access
-        if not interface:
-            interface = self.interface
+        self._emit_signal(interface, name, signature, sigargs, {})
 
-        # convert types of arguments according to signature, using
-        # MethodCallMessage.append(); this will also provide type/length
-        # checks, except for the case of an empty signature
-        if signature == '' and len(args) > 0:
-            raise TypeError('Fewer items found in D-Bus signature than in Python arguments')
-        m = dbus.connection.MethodCallMessage('a.b', '/a', 'a.b', 'a')
-        m.append(signature=signature, *args)
-        args = m.get_args_list()
+    @dbus.service.method(MOCK_IFACE,
+                         in_signature='sssava{sv}',
+                         out_signature='')
+    def EmitSignalDetailed(self, interface: str, name: str, signature: str, sigargs: Tuple[Any, ...], details: PropsType) -> None:
+        '''Emit a signal from the object with extra details.
 
-        fn = lambda self, *args: self.log(f'emit {self.path} {interface}.{name}{_format_args(args)}')
-        fn.__name__ = str(name)
-        dbus_fn = dbus.service.signal(interface)(fn)
-        dbus_fn._dbus_signature = signature
-        dbus_fn._dbus_args = [f'arg{i}' for i in range(1, len(args) + 1)]
-
-        dbus_fn(self, *args)
+        interface: D-Bus interface to send the signal from. For convenience you
+                   can specify '' here to add the method to the object's main
+                   interface (as specified on construction).
+        name: Name of the signal
+        signature: Signature of input arguments; for example "ias" for a signal
+                that takes an int32 and a string array as arguments; see
+                http://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-signatures
+        args: variant array with signal arguments; must match order and type in
+              "signature"
+        details: dictionary with a string key/value entries. Supported keys are:
+            "destination": for the signal destination
+        '''
+        self._emit_signal(interface, name, signature, sigargs, details)
 
     @dbus.service.method(MOCK_IFACE,
                          in_signature='',
@@ -673,38 +716,35 @@ class DBusMockObject(dbus.service.Object):  # pylint: disable=too-many-instance-
                                            'oas', [dbus.ObjectPath(path),
                                                    objects[path].props])
 
-    def mock_method(self, interface: str, dbus_method: str, in_signature: str, *args, **_) -> Any:
+    def mock_method(self, interface: str, dbus_method: str, in_signature: str, *m_args, **_) -> Any:
         '''Master mock method.
 
         This gets "instantiated" in AddMethod(). Execute the code snippet of
         the method and return the "ret" variable if it was set.
         '''
-        # print('mock_method', dbus_method, self, in_signature, args, kwargs, file=sys.stderr)
+        # print('mock_method', dbus_method, self, in_signature, args, _, file=sys.stderr)
 
-        # convert types of arguments according to signature, using
-        # MethodCallMessage.append(); this will also provide type/length
-        # checks, except for the case of an empty signature
-        if in_signature == '' and len(args) > 0:
-            raise TypeError('Fewer items found in D-Bus signature than in Python arguments')
-        m = dbus.connection.MethodCallMessage('a.b', '/a', 'a.b', 'a')
-        m.append(signature=in_signature, *args)
-        args = m.get_args_list()
+        try:
+            args = _convert_args(in_signature, m_args)
 
-        self.log(dbus_method + _format_args(args))
-        self.call_log.append((int(time.time()), str(dbus_method), args))
-        self.MethodCalled(dbus_method, args)
+            self.log(dbus_method + _format_args(args))
+            self.call_log.append((int(time.time()), str(dbus_method), args))
+            self.MethodCalled(dbus_method, args)
 
-        # The code may be a Python 3 string to interpret, or may be a function
-        # object (if AddMethod was called from within Python itself, rather than
-        # over D-Bus).
-        code = self.methods[interface][dbus_method][2]
-        if code and isinstance(code, types.FunctionType):
-            return code(self, *args)
-        if code:
-            loc = locals().copy()
-            exec(code, globals(), loc)  # pylint: disable=exec-used
-            if 'ret' in loc:
-                return loc['ret']
+            # The code may be a Python 3 string to interpret, or may be a function
+            # object (if AddMethod was called from within Python itself, rather than
+            # over D-Bus).
+            code = self.methods[interface][dbus_method][2]
+            if code and isinstance(code, types.FunctionType):
+                return code(self, *args)
+            if code:
+                loc = locals().copy()
+                exec(code, globals(), loc)  # pylint: disable=exec-used
+                if 'ret' in loc:
+                    return loc['ret']
+        except Exception as e:
+            self.log(dbus_method + ' raised: ' + str(e))
+            raise e
 
         return None
 
